//--------------------------------------------------------------------------------
//  Windows bitmap data operations
//    Copyright (c) kei mesuda 2007 
//    (http://www.yomogi.sakura.ne.jp/~si)
//--------------------------------------------------------------------------------
private import std.stdio;
private import std.file;
private import util.color;


// Bitmap File/Info Header 
//--------------------------------------------------
struct BFH {
    char[2] bfType = "BM";
    uint    bfSize = 14;
    ushort  bfReserved1;
    ushort  bfReserved2;
    uint    bfOffBits;
}

struct BIH {
    uint    biSize = 40;
    int     biWidth;
    int     biHeight;
    ushort  biPlanes;
    ushort  biBitCount;
    uint    biCompression;
    uint    biSizeImage;
    int     biXPelsPerMeter;
    int     biYPelsPerMeter;
    uint    biClrUsed;
    uint    biClrImportant;
} 


// load bmp file by specifyed bit size (8,24,32). 
bool loadBitmap(out ubyte[] pix_out, out int width, out int height, char[] nameFile, int bit, bool isRGBA)
{
    int bmpbit;
    scope ubyte[] pxl;
    scope ubyte[] tbl;
    if (readBitmap(nameFile, width, height, bmpbit, pxl, tbl) == null) return false;
    switch (bit) {
    case 8:  return convertBitmap8 (pix_out, pxl, tbl, bmpbit);
    case 24: return convertBitmap24(pix_out, pxl, tbl, bmpbit, isRGBA);
    case 32: return convertBitmap32(pix_out, pxl, tbl, bmpbit, isRGBA);
    default: return false;
    }
}


// read limited information of bitmap
ubyte[] readBitmap(char[] nameFile, out int width, out int height, out int bit, out ubyte[] pixel, out ubyte[] tableColor)
{
    ubyte[] data;
    try data = cast(ubyte[])std.file.read(nameFile);
    catch (Exception ex) throw new Exception(nameFile ~ "; not exist");

    // read header
    if (data.length < 54) throw new Exception(nameFile ~ "; not bitmap file");
    ubyte[] header = data[0..14];
    ubyte[] info   = data[14..54];
    if (header[0]!='B' || header[1]!='M')    throw new Exception(nameFile ~ "; not bitmap file");
    if ((cast(BIH*)info).biCompression != 0) throw new Exception(nameFile ~ "; complessed");

    // get information
    width  = (cast(BIH*)info).biWidth;
    height = (cast(BIH*)info).biHeight;
    bit    = (cast(BIH*)info).biBitCount;

    // refer color table
    uint cntColor = 0;
    if (bit==8 || bit==1) {
        cntColor = (cast(BIH*)info).biClrUsed;
        if (cntColor==0) cntColor=1<<bit;
        if (data.length < 54+cntColor) throw new Exception(nameFile ~ "; color table is broken");
        tableColor = data[54..54+(cntColor<<2)];
    }

    pixel = data[54+(cntColor<<2)..length];
    return data;
}


// convert
bool convertBitmap8(out ubyte[] pix8, ubyte[] pixel, ubyte[] tableColor, int bit)
{
    uint  add, i;
    ubyte f;
    ubyte[256] gray;
    
    switch (bit) {
    case 24:
        pix8.length = pixel.length / 3;
        for (i=0, add=0; i<pix8.length; i++, add+=3) {
            pix8[i] = rgb2gray(pixel[add..add+3]);
        }
        return true;
    case 8:
        pix8.length = pixel.length;
        add=0;
        for (i=0; i<tableColor.length; i+=4) {
            gray[i] = rgb2gray(tableColor[i..i+3]);
        }
        foreach(p; pixel) { 
            pix8[i] = gray[p];
        }
        return true;
    case 1:
        pix8.length = pixel.length * 8;
        add = 0;
        gray[0] = rgb2gray(tableColor[0..3]);
        gray[1] = rgb2gray(tableColor[3..6]);
        foreach(p; pixel) {
            for (i=0; i<8; i++, p<<=1) {
                f=(p&128)>>7;
                pix8[add++] = gray[f];
            }
        }
        return true;
    default:
        return false;
    }
}


// convert
bool convertBitmap24(out ubyte[] pix24, ubyte[] pixel, ubyte[] tableColor, int bit, bool isBGR)
{
    uint  add, i;
    ubyte f;
    
    ubyte[] function(ubyte[], ubyte[]) copybits = (isBGR) ? (&copyLE3) : (&copyBE3);
    
    switch (bit) {
    case 24:
        pix24.length = pixel.length;
        for (i=0; i<pixel.length; i+=3) {
            copybits(pix24[i..i+3], pixel[i..i+3]);
        }
        return true;
    case 8:
        pix24.length = pixel.length * 3;
        add = 0;
        foreach(p; pixel) {
            copybits(pix24[add..add+3], tableColor[p<<2..(p<<2)+3]);
            add+=3;
        }
        return true;
    case 1:
        pix24.length = pixel.length * 24;
        add = 0;
        foreach(p; pixel) {
            for (i=0; i<8; i++, add+=3, p<<=1) {
                f = (p&128)>>5;
                copybits(pix24[add..add+3], tableColor[f..f+3]);
            }
        }
        return true;
    default:
        return false;
    }
}


// convert
bool convertBitmap32(out ubyte[] pix32, ubyte[] pixel, ubyte[] tableColor, int bit, bool isBGRA)
{
    uint  add, i;
    ubyte f;
    
    ubyte[] function(ubyte[], ubyte[]) copybits = (isBGRA) ? (&copyLE3) : (&copyBE3);

    switch (bit) {
    case 24:
        pix32.length = (pixel.length * 4) / 3;
        add = 0;
        for (i=0; i<pix32.length; i+=4, add+=3) {
            copybits(pix32[i..i+3], pixel[add..add+3]);
            pix32[i+3] = 255;
        }
        return true;
    case 8:
        pix32.length = pixel.length * 4;
        add = 0;
        foreach(p; pixel) {
            copybits(pix32[add..add+3], tableColor[p<<2..(p<<2)+3]);
            pix32[add+3] = 255; //tableColor[p+3];
            add+=4;
        }
        return true;
    case 1:
        pix32.length = pixel.length * 32;
        add = 0;
        foreach(p; pixel) {
            for (i=0; i<8; i++, add+=4, p<<=1) {
                f = (p&128)>>5;
                copybits(pix32[add..add+3], tableColor[f..f+3]);
                pix32[add+3] = 255; //tableColor[f+3];
            }
        }
        return true;
    default:
        return false;
    }
}


// alpha color
void transparentAlphaColor(inout ubyte[] pix32, uint colTrans)
{
    colTrans |= 0xff000000;
    for (int i=0; i<pix32.length; i+=4) {
        pix32[i+3] = (rgb2color(pix32[i..i+3]) == colTrans) ? 0 : 255;
    }
}


// alpha map
void setAlphaMap(inout ubyte[] pix32, ubyte[] pix8_alpha)
{
    if (pix32.length != (pix8_alpha.length<<2)) throw new Exception("Alpha map size is not matched");
    for (int i=0; i<pix8_alpha.length; i++) { pix32[(i<<2)+3] = pix8_alpha[i]; }
}

