//------------------------------------------------------------------------------------------
// OpenGL expansion
//------------------------------------------------------------------------------------------
private import std.math;
private import opengl;
private import openglu;
private import util.math;
private import util.color;
private import util.vector;




// set blend function
void setAlphaBlend()  { glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); }
void setAddBlend()    { glBlendFunc(GL_SRC_ALPHA, GL_ONE); }
void setMultBlend()   { glBlendFunc(GL_ZERO, GL_SRC_COLOR); }
void setScreenBlend() { glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ONE); }
void setXorBlend()    { glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ONE_MINUS_SRC_COLOR); }
void setInvertBlend() { glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ZERO); }


// set typical attributions
// Model = enable(depth_test, texture, lighting)
void setModelAttribution()
{
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_BLEND);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_LIGHTING);
}

// Solid = enable(depth_test, lighting)
void setSolidAttribution()
{
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_LIGHTING);
}

// Sprite = enable(blend, texture)
void setSpriteAttribution()
{
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
}


// wrapping of glFog()
void setFog(float start, float end, uint color, float density = 1, int mode = GL_LINEAR)
{
    vector4 col;
    setUint(&col, color);
    glFogi(GL_FOG_MODE,    mode);
    glFogf(GL_FOG_DENSITY, density);
    glFogf(GL_FOG_START,   start);
    glFogf(GL_FOG_END,     end);
    glFogfv(GL_FOG_COLOR,  &col.x);
}


// glGenLists() -> glNewList(GL_COMPILE) -> glEndList()
int composeDisplayList()
{
    int idx = glGenLists(1);
    glNewList(idx, GL_COMPILE);
    return idx;
}

void composeEnd()
{
    glEndList();
}


// pop projectoin matrix
void popScreen()
{
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
}

// set screen size
void setScreen(float fWidth, float fHeight, float fZnear, float fZfar)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    float w_half = fWidth * 0.5;
    float h_half = fHeight * 0.5;
    glOrtho(-w_half, w_half, -h_half, h_half, fZnear, fZfar);
    glMatrixMode(GL_MODELVIEW);
}


// push screen size
void pushScreen(float fWidth, float fHeight)
{
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    float w_half = fWidth * 0.5;
    float h_half = fHeight * 0.5;
    glOrtho(-w_half, w_half, -h_half, h_half, -100, 100);
    glMatrixMode(GL_MODELVIEW);
}


// set screen size with perspective camera, return camera z position
float setScreenPers(float fPersAngle, float fWidth, float fHeight, float fZnear, float fZfar, matrix4* mat=null)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    float dist = fWidth * 0.5 / tan(degree2radian(fPersAngle) * 0.5);
    float ratio = (dist + fZnear) / dist * 0.5;
    float w_half = fWidth * ratio;
    float h_half = fHeight * ratio;

    glFrustum(-w_half, w_half, -h_half, h_half, dist + fZnear, dist + fZfar);
    glTranslatef(0, 0, -dist);
    glMatrixMode(GL_MODELVIEW);
    
    if (mat != null) {
        loadFrustum(mat, -w_half, w_half, -h_half, h_half, dist + fZnear, dist + fZfar);
    }
    
    return dist;
}


// set screen size with frustum camera.
void setScreenFrustum(float fZcamera, float x, float y, float fWidth, float fHeight, float fZnear, float fZfar)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    float r = (fZcamera - fZnear) / fZcamera;
    glFrustum(x*r, y*r, (x+fWidth)*r, (y+fHeight)*r, fZnear - fZcamera, fZfar - fZcamera);
    glTranslatef(0, 0, fZcamera);
    glMatrixMode(GL_MODELVIEW);
}


// set screen size
void setScreenOrtho(float x, float y, float fWidth, float fHeight, float fZnear, float fZfar)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(x, x+fWidth, y, y+fHeight, fZnear, fZfar);
    glMatrixMode(GL_MODELVIEW);
}



